/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    control::ArmarXObjects::controller_creator
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2023
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#include "Component.h"
#include <Eigen/Core>
// #include <SimoxUtility/color/Color.h>
#include "ArmarXCore/core/time/Metronome.h"
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/time/Clock.h>
#include <ArmarXCore/libraries/DecoupledSingleComponent/Decoupled.h>
// for kinesthetic teaching
#include <ArmarXCore/core/csv/CsvWriter.h>
#include <ArmarXGui/interface/StaticPlotterInterface.h>
#include <RobotAPI/libraries/core/math/TimeSeriesUtils.h>
#include <RobotAPI/libraries/core/math/LinearizeAngularTrajectory.h>
#include <RobotAPI/interface/units/RobotUnit/NjointZeroTorqueController.h>
#include <armarx/control/common/utils.h>


namespace armarx::control::components::controller_creator
{

    const std::string
    Component::defaultName = "controller_creator";

    Component::Component()
    {
        addPlugin(m_virtualRobotReaderPlugin);
    }

    Component::~Component() = default;



    armarx::PropertyDefinitionsPtr
    Component::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr def = new armarx::ComponentPropertyDefinitions(getConfigIdentifier());

        def->optional(properties.robotName, "robotName");
        def->optional(properties.robotUnitName, "robotUnitName");

        def->optional(properties.defaultRobotNodeSetLeft, "default_robotNodeSet_left", "default kinematic chain name of left arm.");
        def->optional(properties.defaultRobotNodeSetRight, "default_robotNodeSet_right", "default kinematic chain name of right arm.");

        def->component(m_robot_unit, properties.robotUnitName);
        def->component(m_force_torque_observer, properties.ftObserverName);
        // Publish to a topic (passing the TopicListenerPrx).
        // def->topic(myTopicListener);

        // Subscribe to a topic (passing the topic name).
        // def->topic<PlatformUnitListener>("MyTopic");

        // Use (and depend on) another component (passing the ComponentInterfacePrx).
        // def->component(myComponentProxy)


        // Add a required property. (The component won't start without a value being set.)
        // def->required(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");

        // Add an optional property.
        // def->optional(properties.boxLayerName, "p.box.LayerName", "Name of the box layer in ArViz.");
        // def->optional(properties.numBoxes, "p.box.Number", "Number of boxes to draw in ArViz.");

        return def;
    }


    void
    Component::onInitComponent()
    {
        // Topics and properties defined above are automagically registered.

        // Keep debug observer data until calling `sendDebugObserverBatch()`.
        // (Requires the armarx::DebugObserverComponentPluginUser.)
        // setDebugObserverBatchModeEnabled(true);

    }


    void
    Component::onConnectComponent()
    {
        // Do things after connecting to topics and components.
        ARMARX_CHECK_NOT_NULL(m_robot_unit);

        /* (Requires the armarx::DebugObserverComponentPluginUser.)
        // Use the debug observer to log data over time.
        // The data can be viewed in the ObserverView and the LivePlotter.
        // (Before starting any threads, we don't need to lock mutexes.)
        {
            setDebugObserverDatafield("numBoxes", properties.numBoxes);
            setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
            sendDebugObserverBatch();
        }
        */

        /* (Requires the armarx::ArVizComponentPluginUser.)
        // Draw boxes in ArViz.
        // (Before starting any threads, we don't need to lock mutexes.)
        drawBoxes(properties, arviz);
        */

        /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
        // Setup the remote GUI.
        {
            createRemoteGuiTab();
            RemoteGui_startRunningTask();
        }
        */
        ARMARX_INFO << properties.robotName;
        // m_robot = m_virtualRobotReaderPlugin->get().getRobot(properties.robotName);
        if (!m_robot)
        {
            //            m_robot = RemoteRobot::createLocalCloneFromFile(robotStateComponent, VirtualRobot::RobotIO::eStructure);
            m_robot = addRobot("controller_creator_robot", VirtualRobot::RobotIO::eStructure);
        }
        ARMARX_CHECK_NOT_NULL(m_robot);
        ARMARX_CHECK(synchronizeLocalClone(m_robot));

        m_rns_l = m_robot->getRobotNodeSet(properties.defaultRobotNodeSetLeft);
        m_rns_r = m_robot->getRobotNodeSet(properties.defaultRobotNodeSetRight);

        m_ft_l.reset(new armarx::control::common::ft::ForceTorqueUtility(m_force_torque_observer, properties.ftSensorNameLeft));
        m_ft_r.reset(new armarx::control::common::ft::ForceTorqueUtility(m_force_torque_observer, properties.ftSensorNameRight));

        ARMARX_IMPORTANT << "controller creator is ready";

    }


    void
    Component::onDisconnectComponent()
    {
    }


    void
    Component::onExitComponent()
    {
    }

    std::string
    Component::createImpedanceController(
            const std::string& controllerNamePrefix,
            const std::string& robotNodeSet,
            const std::string& configFilename)
    {
        /// create controllers with controller builder
        auto builder = controllerBuilder<control::common::ControllerType::TSImpedance>();
        builder.withNodeSet(robotNodeSet).withNamePrefix(controllerNamePrefix);

        /// load predefined config from file
        {
            if (configFilename.empty())
            {
                const std::string controllerClassName =
                        control::common::ControllerTypeNames.to_name(control::common::ControllerType::TSImpedance);
                const armarx::PackagePath configPath("armarx_control", "controller_config/" + controllerClassName + "/default.json");
                ARMARX_IMPORTANT << "Using config `" << configPath.toSystemPath() << "`.";
                ARMARX_CHECK(std::filesystem::exists(configPath.toSystemPath()));

                builder.withConfig(configPath.toSystemPath());
            }
            else
            {
                builder.withConfig(configFilename);
            }


            /// to be removed
            ARMARX_INFO << VAROUT(builder.config().kpImpedance);
            ARMARX_INFO << VAROUT(builder.config().kdImpedance);
            ARMARX_INFO << VAROUT(builder.config().kpNullspace);
            ARMARX_INFO << VAROUT(builder.config().kdNullspace);
            ARMARX_INFO << VAROUT(builder.config().desiredPose);
            ARMARX_INFO << VAROUT(builder.config().desiredTwist);
            if (builder.config().desiredNullspaceJointAngles.has_value())
                ARMARX_INFO << VAROUT(builder.config().desiredNullspaceJointAngles.value());
            ARMARX_INFO << VAROUT(builder.config().torqueLimit);
            ARMARX_INFO << VAROUT(builder.config().qvelFilter);
        }

        ARMARX_CHECK_NOT_NULL(m_robot);
        ARMARX_CHECK(synchronizeLocalClone(m_robot));
        const auto rns = m_robot->getRobotNodeSet(robotNodeSet);


        builder.config().desiredPose = rns->getTCP()->getPoseInRootFrame();
        ARMARX_IMPORTANT << VAROUT(builder.config().desiredPose);

        ARMARX_TRACE;
        ARMARX_INFO << "creating controller";

        auto ctrlWrapper = builder.create();
        const std::string controllerName = builder.getControllerName();
        m_robotNodeSets[controllerName] = rns;
        m_controllerJointNumbers[controllerName] = rns->getJointValues().size();

        // The controller is scoped by default, which means that if the ControllerWrapper goes out of scope, it will deactivate
        // and delete the NJointController. To prevent this, you can use the daemonize() method if needed.
        ctrlWrapper->daemonize(true);

        ARMARX_TRACE;
        ARMARX_CHECK(ctrlWrapper.has_value());
        m_ctrlTSImpedance[controllerName] = std::nullopt;
        m_ctrlTSImpedance[controllerName].emplace(std::move(ctrlWrapper.value()));

        Clock::WaitFor(Duration::MilliSeconds(300));
        ARMARX_INFO << "Activating controller";
        m_ctrlTSImpedance[controllerName]->updateConfig();
        m_ctrlTSImpedance[controllerName]->activate();

        m_prevTargetPose[controllerName] = rns->getTCP()->getPoseInRootFrame();
        m_prevNullspaceJointTarget[controllerName] = rns->getJointValuesEigen();
        return controllerName;
    }

    std::string
    Component::createImpedanceMPController(
            const std::string& controllerNamePrefix,
            const std::string& robotNodeSet,
            const std::string& configFilename)
    {
        /// create controllers with controller builder
        auto builder = controllerBuilder<control::common::ControllerType::TSImpedanceMP>();
        builder.allowReuse(true).withNodeSet(robotNodeSet).withNamePrefix(controllerNamePrefix);

        /// load predefined config from file
        {
            if (configFilename.empty())
            {
                const std::string controllerClassName =
                        control::common::ControllerTypeNames.to_name(control::common::ControllerType::TSImpedanceMP);
                const armarx::PackagePath configPath("armarx_control", "controller_config/" + controllerClassName + "/default.json");
                ARMARX_IMPORTANT << "Using config `" << configPath.toSystemPath() << "`.";
                ARMARX_CHECK(std::filesystem::exists(configPath.toSystemPath()));

                builder.withConfig(configPath.toSystemPath());
            }
            else
            {
                builder.withConfig(configFilename);
            }


            /// to be removed
            ARMARX_INFO << VAROUT(builder.config().kpImpedance);
            ARMARX_INFO << VAROUT(builder.config().kdImpedance);
            ARMARX_INFO << VAROUT(builder.config().kpNullspace);
            ARMARX_INFO << VAROUT(builder.config().kdNullspace);
            ARMARX_INFO << VAROUT(builder.config().desiredPose);
            ARMARX_INFO << VAROUT(builder.config().desiredTwist);
            if (builder.config().desiredNullspaceJointAngles.has_value())
                ARMARX_INFO << VAROUT(builder.config().desiredNullspaceJointAngles.value());
            ARMARX_INFO << VAROUT(builder.config().torqueLimit);
            ARMARX_INFO << VAROUT(builder.config().qvelFilter);
        }

        ARMARX_CHECK_NOT_NULL(m_robot);
        ARMARX_CHECK(synchronizeLocalClone(m_robot));
        const auto rns = m_robot->getRobotNodeSet(robotNodeSet);


        builder.config().desiredPose = rns->getTCP()->getPoseInRootFrame();
        ARMARX_IMPORTANT << VAROUT(builder.config().desiredPose);

        ARMARX_TRACE;
        ARMARX_INFO << "creating controller";

        auto ctrlWrapper = builder.create();
        const std::string controllerName = builder.getControllerName();
        m_robotNodeSets[controllerName] = rns;
        m_controllerJointNumbers[controllerName] = rns->getJointValues().size();

        // The controller is scoped by default, which means that if the ControllerWrapper goes out of scope, it will deactivate
        // and delete the NJointController. To prevent this, you can use the daemonize() method if needed.
        ctrlWrapper->daemonize(true);

        ARMARX_TRACE;
        ARMARX_CHECK(ctrlWrapper.has_value());
        m_ctrlTSImpedanceMP[controllerName] = std::nullopt;
        m_ctrlTSImpedanceMP[controllerName].emplace(std::move(ctrlWrapper.value()));

        Clock::WaitFor(Duration::MilliSeconds(300));
        ARMARX_INFO << "Activating controller";
        m_ctrlTSImpedanceMP[controllerName]->updateConfig();

        /// only for test
        //        auto ctrl = NJointTSImpedanceMPControllerInterfacePrx::checkedCast(m_ctrlTSImpedanceMP[controllerName]->ctrl());
        //        ctrl->learnFromCSV(std::vector<std::string>());

        m_ctrlTSImpedanceMP[controllerName]->activate();

        m_prevTargetPose[controllerName] = rns->getTCP()->getPoseInRootFrame();
        m_prevNullspaceJointTarget[controllerName] = rns->getJointValuesEigen();
        //        /// only for test
        //        auto currentPose = m_prevTargetPose[controllerName];
        //        auto targetPose = currentPose;
        //        targetPose(2, 3) += 300.0;
        //        ctrl->start(
        //                "default",
        //                armarx::control::common::mat4ToDVec(currentPose),
        //                armarx::control::common::mat4ToDVec(targetPose),
        //                -1);

        return controllerName;
    }

    std::string
    Component::createAdmittanceController(
            const std::string &controllerNamePrefix,
            const std::string &robotNodeSet,
            const std::string& configFilename)
    {
        ARMARX_INFO << "creating ts admittance controller";
        /// create controllers with controller builder
        auto builder = controllerBuilder<control::common::ControllerType::TSAdmittance>();
        builder.withNodeSet(robotNodeSet).withNamePrefix(controllerNamePrefix);

        /// load predefined config from file
        {
            if (configFilename.empty())
            {
                const std::string controllerClassName =
                        control::common::ControllerTypeNames.to_name(control::common::ControllerType::TSAdmittance);
                const armarx::PackagePath configPath("armarx_control", "controller_config/" + controllerClassName + "/default.json");
                ARMARX_IMPORTANT << "Using config `" << configPath.toSystemPath() << "`.";
                ARMARX_CHECK(std::filesystem::exists(configPath.toSystemPath()));

                builder.withConfig(configPath.toSystemPath());
            }
            else
            {
                builder.withConfig(configFilename);
            }
        }

        ARMARX_CHECK_NOT_NULL(m_robot);
        ARMARX_CHECK(synchronizeLocalClone(m_robot));

        const auto rns = m_robot->getRobotNodeSet(robotNodeSet);

        builder.config().desiredPose = rns->getTCP()->getPoseInRootFrame();
        ARMARX_IMPORTANT << VAROUT(builder.config().desiredPose);

        ARMARX_TRACE;
        ARMARX_INFO << "creating controller";

        auto ctrlWrapper = builder.create();
        const std::string controllerName = builder.getControllerName();
        m_robotNodeSets[controllerName] = rns;
        m_controllerJointNumbers[controllerName] = rns->getJointValues().size();

        // The controller is scoped by default, which means that if the ControllerWrapper goes out of scope, it will deactivate
        // and delete the NJointController. To prevent this, you can use the daemonize() method if needed.
        ctrlWrapper->daemonize(true);

        ARMARX_TRACE;
        ARMARX_CHECK(ctrlWrapper.has_value());
        m_ctrlTSAdmittance[controllerName] = std::nullopt;
        m_ctrlTSAdmittance[controllerName].emplace(std::move(ctrlWrapper.value()));

        Clock::WaitFor(Duration::MilliSeconds(300));
        ARMARX_INFO << "Activating controller";
        m_ctrlTSAdmittance[controllerName]->updateConfig();
        m_ctrlTSAdmittance[controllerName]->activate();

        m_prevTargetPose[controllerName] = rns->getTCP()->getPoseInRootFrame();
        m_prevNullspaceJointTarget[controllerName] = rns->getJointValuesEigen();
        return controllerName;
    }


    std::string
    Component::createController(
            const std::string& controllerNamePrefix,
            const std::string& robotNodeSet,
            const std::string& controllerType,
            const std::string& configFilename,
            const Ice::Current&)
    {
        ARMARX_INFO << "creating controller of type " << controllerType << " with robot node set " << robotNodeSet;
        ARMARX_INFO << VAROUT(robotNodeSet) << VAROUT(controllerType) << VAROUT(configFilename);
        RobotUnitInterfacePrx robotUnit = getControlComponentPlugin().getRobotUnitPlugin().getRobotUnit();
        robotUnit->loadLibFromPackage("armarx_control", "libarmarx_control_njoint_controller");

        std::string controllerName;

        auto type = control::common::ControllerTypeShort.from_name(controllerType);
        switch (type) {
            case control::common::ControllerType::TSImpedance:
                controllerName = createImpedanceController(controllerNamePrefix, robotNodeSet, configFilename);
                break;
            case control::common::ControllerType::TSAdmittance:
                controllerName = createAdmittanceController(controllerNamePrefix, robotNodeSet, configFilename);
                break;
            case control::common::ControllerType::TSImpedanceMP:
                controllerName = createImpedanceMPController(controllerNamePrefix, robotNodeSet, configFilename);
                break;

            default:
                ARMARX_INFO << "controller type: " << controllerType << " is not supported yet";
        }
        return controllerName;
    }

    bool
    Component::updateTargetPose(
            const std::string &controllerName,
            const std::string& controllerType,
            const Ice::FloatSeq& targetPoseVector,
            const Ice::Current&)
    {
        auto names = control::common::ControllerTypeShort.names();
        const bool is_in = names.find(controllerType) != names.end();
        ARMARX_CHECK_EQUAL(7, targetPoseVector.size());
        if (!is_in)
        {
            ARMARX_ERROR << "controller type not supported";
            return false;
        }
        if (targetPoseVector.size() != 7)
        {
            ARMARX_ERROR << "target pose vector " << targetPoseVector << " should follow (x, y, z, w, rx, ry, rz) convention";
            return false;
        }

        m_prevTargetPose[controllerName] = VirtualRobot::MathTools::quat2eigen4f(targetPoseVector.at(4), targetPoseVector.at(5), targetPoseVector.at(6), targetPoseVector.at(3));
        m_prevTargetPose[controllerName](0, 3) = targetPoseVector.at(0);
        m_prevTargetPose[controllerName](1, 3) = targetPoseVector.at(1);
        m_prevTargetPose[controllerName](2, 3) = targetPoseVector.at(2);

        auto type = control::common::ControllerTypeShort.from_name(controllerType);
        switch (type) {
            case control::common::ControllerType::TSImpedance:
                ARMARX_CHECK(m_ctrlTSImpedance[controllerName]);
                m_ctrlTSImpedance[controllerName]->config.desiredPose = m_prevTargetPose[controllerName];
                m_ctrlTSImpedance[controllerName]->config.desiredNullspaceJointAngles = m_prevNullspaceJointTarget[controllerName];
                m_ctrlTSImpedance[controllerName]->updateConfig();
                break;
            case control::common::ControllerType::TSAdmittance:
                ARMARX_CHECK(m_ctrlTSAdmittance[controllerName]);
                m_ctrlTSAdmittance[controllerName]->config.desiredPose = m_prevTargetPose[controllerName];
                m_ctrlTSAdmittance[controllerName]->config.desiredNullspaceJointAngles = m_prevNullspaceJointTarget[controllerName];
                m_ctrlTSAdmittance[controllerName]->updateConfig();
                break;

            default:
                ARMARX_INFO << "controller type: " << controllerType << " is not supported yet";
                return false;
        }
        return true;
    }

    bool
    Component::updateTargetNullspaceJointAngles(
            const std::string &controllerName,
            const std::string &controllerType,
            const Ice::FloatSeq& targetNullspaceJointAngle,
            const Ice::Current &)
    {
        auto names = control::common::ControllerTypeShort.names();
        const bool is_in = names.find(controllerType) != names.end();
        ARMARX_CHECK_EQUAL(m_controllerJointNumbers[controllerName], targetNullspaceJointAngle.size());
        if (!is_in)
        {
            ARMARX_ERROR << "controller type not supported";
            return false;
        }

        Eigen::VectorXf targetJointAngles = Eigen::VectorXf::Zero(m_controllerJointNumbers[controllerName]);
        for (size_t i = 0; i < m_controllerJointNumbers[controllerName]; i++)
        {
            targetJointAngles(i) = targetNullspaceJointAngle[i];
        }

        auto type = control::common::ControllerTypeShort.from_name(controllerType);
        switch (type) {
            case control::common::ControllerType::TSImpedance:
                ARMARX_CHECK(m_ctrlTSImpedance[controllerName]);
                m_ctrlTSImpedance[controllerName]->config.desiredNullspaceJointAngles = targetJointAngles;
                m_ctrlTSImpedance[controllerName]->updateConfig();
                break;
            case control::common::ControllerType::TSAdmittance:
                ARMARX_CHECK(m_ctrlTSAdmittance[controllerName]);
                m_ctrlTSAdmittance[controllerName]->config.desiredNullspaceJointAngles = targetJointAngles;
                m_ctrlTSAdmittance[controllerName]->updateConfig();
                break;

            default:
                ARMARX_INFO << "controller type: " << controllerType << " is not supported yet";
                return false;
        }
        return true;
    }

    bool
    Component::updateTarget(
            const std::string &controllerName,
            const std::string &controllerType,
            const Ice::FloatSeq& targetPoseVector,
            const Ice::FloatSeq& targetNullspaceJointAngle,
            const Ice::Current &)
    {
        auto names = control::common::ControllerTypeShort.names();
        const bool is_in = names.find(controllerType) != names.end();
        if (!is_in)
        {
            ARMARX_ERROR << "controller type not supported";
            return false;
        }

        bool updataPose = false;
        if (targetPoseVector.size() > 0)
        {
            if (targetPoseVector.size() != 7)
            {
                ARMARX_ERROR << "target pose vector " << targetPoseVector << " should follow (x, y, z, w, rx, ry, rz) convention";
                return false;
            }
            updataPose = true;
            m_prevTargetPose[controllerName] = VirtualRobot::MathTools::quat2eigen4f(targetPoseVector.at(4), targetPoseVector.at(5), targetPoseVector.at(6), targetPoseVector.at(3));
            m_prevTargetPose[controllerName](0, 3) = targetPoseVector.at(0);
            m_prevTargetPose[controllerName](1, 3) = targetPoseVector.at(1);
            m_prevTargetPose[controllerName](2, 3) = targetPoseVector.at(2);
        }

        bool updateNullspace = false;
        if (targetNullspaceJointAngle.size() > 0)
        {
            for (size_t i = 0; i < m_controllerJointNumbers[controllerName]; i++)
            {
                m_prevNullspaceJointTarget[controllerName](i) = targetNullspaceJointAngle[i];
            }
        }


        auto type = control::common::ControllerTypeShort.from_name(controllerType);
        switch (type) {
            case control::common::ControllerType::TSImpedance:
                ARMARX_CHECK(m_ctrlTSImpedance[controllerName]);
                if (updataPose)
                {
                    m_ctrlTSImpedance[controllerName]->config.desiredPose = m_prevTargetPose[controllerName];
                }
                if (updateNullspace)
                {
                    m_ctrlTSImpedance[controllerName]->config.desiredNullspaceJointAngles = m_prevNullspaceJointTarget[controllerName];
                }
                m_ctrlTSImpedance[controllerName]->updateConfig();
                break;
            case control::common::ControllerType::TSAdmittance:
                ARMARX_CHECK(m_ctrlTSAdmittance[controllerName]);
                if (updataPose)
                {
                    m_ctrlTSAdmittance[controllerName]->config.desiredPose = m_prevTargetPose[controllerName];
                }
                if (updateNullspace)
                {
                    m_ctrlTSAdmittance[controllerName]->config.desiredNullspaceJointAngles = m_prevNullspaceJointTarget[controllerName];
                }
                m_ctrlTSAdmittance[controllerName]->updateConfig();
                break;

            default:
                ARMARX_INFO << "controller type: " << controllerType << " is not supported yet";
                return false;
        }
        return true;
    }


    std::string Component::kinestheticTeaching(
            const std::string& kinematicChainName,
            const std::string& side,
            const std::string& fileName,
            const Ice::Current&)
    {
        /// ---------------------------------- creating zero torque controller ----------------------------------
        ARMARX_INFO << "Kinesthetic teaching: Initialize zero torque controllers for " << VAROUT(kinematicChainName);
        const std::string ctrl_class_name = "NJointZeroTorqueController";
        const std::string ctrl_name = getDefaultName() + "_kinesthetic_teaching_" + ctrl_class_name;

        NJointZeroTorqueControllerConfigPtr config(new NJointZeroTorqueControllerConfig());
        config->maxTorque = 0;

        auto nodeSet = m_robot->getRobotNodeSet(kinematicChainName);
        for (auto& name : nodeSet->getNodeNames())
        {
            config->jointNames.push_back(name);
        }
        NJointZeroTorqueControllerInterfacePrx zeroTorqueController
                = NJointZeroTorqueControllerInterfacePrx::checkedCast(
                          m_robot_unit->createNJointController(
                              ctrl_class_name,
                              ctrl_name,
                              config));
        ARMARX_CHECK_NOT_NULL(zeroTorqueController);
        ARMARX_INFO << "zero torque controller created";

        float motionThresholdRadian = 0.1f;
        float noMotionTimeoutMs = 2000.0f;

        VirtualRobot::RobotNodeSetPtr rns = m_robot->getRobotNodeSet(kinematicChainName);
        std::vector<std::string> jointNames = rns->getNodeNames();
        IceUtil::Time start = TimeUtil::GetTime();
        IceUtil::Time lastMotionTime = start;
        bool initialMotionDetected = false;

        /// ---------------------------------- prepare data containers ----------------------------------
        /// recording data
        std::vector<std::vector<float>> jointData;
        std::vector<std::vector<float>> cartData;
        std::vector<float> timestamps;

        /// init data
        Eigen::VectorXf jve = rns->getJointValuesEigen();
        Eigen::Matrix4f initPose = rns->getTCP()->getPoseInRootFrame();
        VirtualRobot::MathTools::Quaternion initQuat = VirtualRobot::MathTools::eigen4f2quat(initPose);
        VirtualRobot::MathTools::Quaternion oldquatE = initQuat;

        /// joint space
        for (size_t i = 0; i < jointNames.size(); i++)
        {
            std::vector<float> cjoint {};
            // cjoint.push_back(jve(i));
            jointData.push_back(cjoint);
        }

        /// task space
        float quatvec[] = {initPose(0, 3), initPose(1, 3), initPose(2, 3), initQuat.w, initQuat.x, initQuat.y,  initQuat.z};
        for (size_t i = 0; i < 7; ++i)
        {
            std::vector<float> cpos {};
            // cpos.push_back(quatvec[i]);
            cartData.push_back(cpos);
        }


        int motionStartIndex = 0;
        int motionEndIndex = 0;

        int sigmaMs = 1;
        int sampleMs = 10;

        float forceSpike = 0.0f;
        Eigen::Vector3f initForce;
        if (side == "right")
        {
            initForce = m_ft_r->getForce();
        }
        else
        {
            initForce = m_ft_l->getForce();
        }
        ARMARX_IMPORTANT << "=============== Waiting for force spike ===============";
        while (forceSpike < 10.0f)
        {
            Eigen::Vector3f force;
            if (side == "right")
            {
                force = m_ft_r->getForce();
            }
            else
            {
                force = m_ft_l->getForce();
            }
            forceSpike = (force - initForce).norm();
        }
        ARMARX_INFO << "Start the demonstration ...";

        zeroTorqueController->activateController();
        while (true) // stop run function if returning true
        {
            synchronizeLocalClone(m_robot);

            IceUtil::Time now = TimeUtil::GetTime();

            Eigen::VectorXf jve2 = rns->getJointValuesEigen();
            if ((jve2 - jve).norm() > motionThresholdRadian)
            {
                if (!initialMotionDetected)
                {
                    ARMARX_IMPORTANT << "Start recording";
                    motionStartIndex = 0;
                    //start = now;
                }
                jve = jve2;
                initialMotionDetected = true;
                lastMotionTime = now;
            }

            double t = (now - start).toSecondsDouble();

            std::vector<float> jvs = rns->getJointValues();
            for (size_t i = 0; i < jvs.size(); i++)
            {
                jointData.at(i).push_back(jvs.at(i));
            }

            Eigen::Matrix4f tcpPose = rns->getTCP()->getPoseInRootFrame();
            cartData.at(0).push_back(tcpPose(0, 3));
            cartData.at(1).push_back(tcpPose(1, 3));
            cartData.at(2).push_back(tcpPose(2, 3));
            VirtualRobot::MathTools::Quaternion quat = VirtualRobot::MathTools::eigen4f2quat(tcpPose);

            // check flipped orientation quaternion
            double dotProduct = quat.w * oldquatE.w + quat.x * oldquatE.x + quat.y * oldquatE.y + quat.z * oldquatE.z;
            if (dotProduct < 0)
            {
                cartData.at(3).push_back(-quat.w);
                cartData.at(4).push_back(-quat.x);
                cartData.at(5).push_back(-quat.y);
                cartData.at(6).push_back(-quat.z);
                oldquatE.w = -quat.w;
                oldquatE.x = -quat.x;
                oldquatE.y = -quat.y;
                oldquatE.z = -quat.z;

            }
            else
            {
                cartData.at(3).push_back(quat.w);
                cartData.at(4).push_back(quat.x);
                cartData.at(5).push_back(quat.y);
                cartData.at(6).push_back(quat.z);

                oldquatE.w = quat.w;
                oldquatE.x = quat.x;
                oldquatE.y = quat.y;
                oldquatE.z = quat.z;
            }

            timestamps.push_back(t);

            if (initialMotionDetected && (now - lastMotionTime).toMilliSeconds() > noMotionTimeoutMs)
            {
                ARMARX_IMPORTANT << "Stop recording";
                motionEndIndex = timestamps.size();// - noMotionTimeoutMs / sampleMs;
                break;
            }

            TimeUtil::USleep(sampleMs * 1000);
        }
        ARMARX_IMPORTANT << "Demonstration done! Processing data ...";

        zeroTorqueController->deactivateController();
        while (zeroTorqueController->isControllerActive())
        {
            usleep(10000);
        }
        zeroTorqueController->deleteController();
        enableVelocityMode();

        motionStartIndex -= sigmaMs / sampleMs * 2;
        if (motionStartIndex < 0)
        {
            motionStartIndex = 0;
        }
        //motionEndIndex += sigmaMs / sampleMs * 2;
        //if (motionEndIndex >= data.size())
        //{
        //    motionEndIndex = data.size() - 1;
        //}

        ARMARX_IMPORTANT << VAROUT(motionStartIndex) << VAROUT(motionEndIndex);

        ARMARX_IMPORTANT << "copy timestamps";
        timestamps = std::vector<float>(timestamps.begin() + motionStartIndex, timestamps.begin() + motionEndIndex);
        for (size_t n = 1; n < timestamps.size(); n++)
        {
            timestamps.at(n) = timestamps.at(n) - timestamps.at(0);
        }
        timestamps.at(0) = 0;

        ARMARX_IMPORTANT << VAROUT(timestamps);
        std::vector<float> newT = math::TimeSeriesUtils::MakeTimestamps(timestamps.at(0), timestamps.at(timestamps.size() - 1), timestamps.size());
        ARMARX_IMPORTANT << VAROUT(newT);

        {
            StringFloatSeqDict plotFiltered;
            StringFloatSeqDict plotOrig;
            StringFloatSeqDict plotResampled;
            for (size_t i = 0; i < jointData.size(); i++)
            {
                jointData.at(i) = std::vector<float>(jointData.at(i).begin() + motionStartIndex, jointData.at(i).begin() + motionEndIndex);
                if (rns->getNode(i)->isLimitless())
                {
                    math::LinearizeAngularTrajectory::LinearizeRef(jointData.at(i));
                }
                plotOrig[jointNames.at(i)] = jointData.at(i);
                jointData.at(i) = math::TimeSeriesUtils::Resample(timestamps, jointData.at(i), newT);
                plotResampled[jointNames.at(i)] = jointData.at(i);
                jointData.at(i) = math::TimeSeriesUtils::ApplyGaussianFilter(jointData.at(i), sigmaMs * 0.001f, sampleMs * 0.001f, math::TimeSeriesUtils::BorderMode::Nearest);
                plotFiltered[jointNames.at(i)] = jointData.at(i);
            }
            //            getStaticPlotter()->addPlotWithTimestampVector("original joint motion", timestamps, plotOrig);
            //            getStaticPlotter()->addPlotWithTimestampVector("resampled joint motion", newT, plotResampled);
            //            getStaticPlotter()->addPlotWithTimestampVector("filtered joint motion", newT, plotFiltered);
        }

        std::vector<std::string> cartName = {"x", "y", "z", "qw", "qx", "qy", "qz"};

        timestamps = newT;

        std::string trajName;
        {
            IceUtil::Time now = armarx::rtNow();
            time_t timer = now.toSeconds();
            struct tm* ts;
            char buffer[80];
            ts = localtime(&timer);
            strftime(buffer, 80, "%Y-%m-%d-%H-%M-%S", ts);
            std::string str(buffer);
            trajName = fileName + "-" + str;
        }


        double cutoutTime = (noMotionTimeoutMs / 1000.0f) * 0.8;

        std::string packageName = "armarx_control";

        armarx::CMakePackageFinder finder(packageName);
        std::string dataDir = finder.getDataDir() + "/" + packageName + "/kinesthetic_teaching/";
        std::string returnFileNamePattern = dataDir + trajName;

        /// write task space data to RobotControllers Data dir
        ARMARX_IMPORTANT << "writing task space data!";
        std::string dmpForwardFile;
        std::string dmpBackwardFile;
        {
            std::vector<std::string> header;
            header.push_back("timestamp");
            for (size_t i = 0; i < cartName.size(); ++i)
            {
                header.push_back(cartName.at(i));
            }

            std::string ffname = trajName + "-ts-forward.csv";
            std::string bfname = trajName + "-ts-backward.csv";
            dmpForwardFile = dataDir + ffname;
            dmpBackwardFile = dataDir + bfname;
            std::vector<std::string> fflist;
            fflist.push_back(ffname);

            std::vector<std::string> bflist;
            bflist.push_back(bfname);

            CsvWriter fwriter(dmpForwardFile, header, false);
            CsvWriter bwriter(dmpBackwardFile, header, false);

            ARMARX_IMPORTANT << "writing taskspace trajectory as " << ffname << " and " << bfname;

            double endTime = timestamps.at(timestamps.size() - 1);
            for (size_t n = 0; n < timestamps.size(); n++)
            {

                std::vector<float> row;
                row.push_back(timestamps.at(n));
                for (size_t i = 0; i < cartData.size(); i++)
                {
                    row.push_back(cartData.at(i).at(n));
                }
                fwriter.write(row);

                if (endTime - timestamps.at(n) < cutoutTime)
                {
                    Eigen::Matrix4f endPose = VirtualRobot::MathTools::quat2eigen4f(row.at(5), row.at(6), row.at(7), row.at(4));
                    endPose(0, 3) = row.at(1);
                    endPose(1, 3) = row.at(2);
                    endPose(2, 3) = row.at(3);
                    break;
                }
            }
            fwriter.close();
            int nt = 0;
            for (int n = timestamps.size() - 1; n >= 0 && nt < static_cast<int>(timestamps.size()); n--, nt++)
            {
                std::vector<float> row;
                row.push_back(timestamps.at(nt));
                for (size_t i = 0; i < cartData.size(); i++)
                {
                    row.push_back(cartData.at(i).at(n));
                }
                bwriter.write(row);

                if (n == 0)
                {
                    Eigen::Matrix4f endPose = VirtualRobot::MathTools::quat2eigen4f(row.at(5), row.at(6), row.at(7), row.at(4));
                    endPose(0, 3) = row.at(1);
                    endPose(1, 3) = row.at(2);
                    endPose(2, 3) = row.at(3);
                    break;
                }
            }
            bwriter.close();
        }

        /// write joint space data to RobotControllers Data dir
        ARMARX_IMPORTANT << "writing joint space data!";
        std::string dmpJSForwardFile;
        std::string dmpJSBackwardFile;
        {
            std::vector<std::string> header;
            header.push_back("timestamp");
            for (size_t i = 0; i < jointNames.size(); ++i)
            {
                header.push_back(jointNames.at(i));
            }

            std::string ffname = trajName + "-js-forward.csv";
            std::string bfname = trajName + "-js-backward.csv";
            dmpJSForwardFile = dataDir + ffname;
            dmpJSBackwardFile = dataDir + bfname;
            std::vector<std::string> fflist;
            fflist.push_back(ffname);

            std::vector<std::string> bflist;
            bflist.push_back(bfname);

            CsvWriter fwriter(dmpJSForwardFile, header, false);
            CsvWriter bwriter(dmpJSBackwardFile, header, false);

            ARMARX_IMPORTANT << "writing jointspace trajectory as " << ffname << " and " << bfname;

            double endTime = timestamps.at(timestamps.size() - 1);
            for (size_t n = 0; n < timestamps.size(); n++)
            {

                std::vector<float> row;
                row.push_back(timestamps.at(n));
                for (size_t i = 0; i < jointNames.size(); i++)
                {
                    row.push_back(jointData.at(i).at(n));
                }
                fwriter.write(row);

                if (endTime - timestamps.at(n) < cutoutTime)
                {
                    std::vector<float> endJointValue;
                    for (size_t i = 0; i < jointNames.size(); i++)
                    {
                        endJointValue.push_back(jointData.at(i).at(n));
                    }
                    break;
                }
            }
            fwriter.close();
            ARMARX_INFO << "forward joint space data written to fine!";
            int nt = 0;
            for (int n = timestamps.size() - 1; n >= 0 && nt < static_cast<int>(timestamps.size()); n--, nt++)
            {
                std::vector<float> row;
                row.push_back(timestamps.at(nt));
                for (size_t i = 0; i < jointNames.size(); i++)
                {
                    row.push_back(jointData.at(i).at(n));
                }
                bwriter.write(row);

                if (n == 0)
                {
                    std::vector<float> endJointValue;
                    for (size_t i = 0; i < jointNames.size(); i++)
                    {
                        endJointValue.push_back(jointData.at(i).at(n));
                    }
                    break;
                }
            }
            bwriter.close();
        }
        return returnFileNamePattern;
//        return "";
    }

    void
    Component::enableVelocityMode(const Ice::Current&)
    {
        ARMARX_INFO << "enable velocity mode";
        // left arm
        {
            std::vector<std::string> joint_names = m_rns_l->getNodeNames();
            NameValueMap target_vels;
            NameControlModeMap control_modes;
            for (const std::string& joint_name : joint_names)
            {
                control_modes[joint_name] = eVelocityControl;
                target_vels[joint_name] = 0;
            }
        }
        // right arm
        {
            std::vector<std::string> joint_names = m_rns_r->getNodeNames();
            NameValueMap target_vels;
            NameControlModeMap control_modes;
            for (const std::string& joint_name : joint_names)
            {
                control_modes[joint_name] = eVelocityControl;
                target_vels[joint_name] = 0;
            }
        }
//        // platform
//        {
//            m_platform->move(0.0f, 0.0f, 0.0f);
//        }
    }


    Ice::FloatSeq
    Component::getPoseInRootFrame(const std::string& nodeName, const Ice::Current&)
    {
        ARMARX_CHECK(synchronizeLocalClone(m_robot));
        Eigen::Matrix4f pose = m_robot->getRobotNode(nodeName)->getPoseInRootFrame();
        ARMARX_INFO << VAROUT(pose);
        std::vector<Ice::Float> pose_seq;
        for (int i = 0; i < pose.rows(); i++){
            for (int j = 0; j < pose.cols(); j++){
                pose_seq.push_back(pose(i, j));
            }
        }
        return pose_seq;
    }

    Ice::FloatSeq
    Component::getPrevTargetPose(
            const std::string &controllerName,
            const Ice::Current&)
    {
        std::vector<Ice::Float> pose_seq;
        for (int i = 0; i < m_prevTargetPose[controllerName].rows(); i++){
            for (int j = 0; j < m_prevTargetPose[controllerName].cols(); j++){
                pose_seq.push_back(m_prevTargetPose[controllerName](i, j));
            }
        }
        return pose_seq;
    }

    Ice::FloatSeq
    Component::getPrevNullspaceTargetAngle(
            const std::string &controllerName,
            const Ice::Current&)
    {
        std::vector<Ice::Float> angle_seq;
        for (int i = 0; i < m_prevNullspaceJointTarget[controllerName].rows(); i++){
            angle_seq.push_back(m_prevNullspaceJointTarget[controllerName](i));
        }
        return angle_seq;
    }


    std::string
    Component::getDefaultName() const
    {
        return Component::defaultName;
    }


    std::string
    Component::GetDefaultName()
    {
        return Component::defaultName;
    }


    /* (Requires the armarx::LightweightRemoteGuiComponentPluginUser.)
    void
    Component::createRemoteGuiTab()
    {
        using namespace armarx::RemoteGui::Client;

        // Setup the widgets.

        tab.boxLayerName.setValue(properties.boxLayerName);

        tab.numBoxes.setValue(properties.numBoxes);
        tab.numBoxes.setRange(0, 100);

        tab.drawBoxes.setLabel("Draw Boxes");

        // Setup the layout.

        GridLayout grid;
        int row = 0;
        {
            grid.add(Label("Box Layer"), {row, 0}).add(tab.boxLayerName, {row, 1});
            ++row;

            grid.add(Label("Num Boxes"), {row, 0}).add(tab.numBoxes, {row, 1});
            ++row;

            grid.add(tab.drawBoxes, {row, 0}, {2, 1});
            ++row;
        }

        VBoxLayout root = {grid, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void
    Component::RemoteGui_update()
    {
        if (tab.boxLayerName.hasValueChanged() || tab.numBoxes.hasValueChanged())
        {
            std::scoped_lock lock(propertiesMutex);
            properties.boxLayerName = tab.boxLayerName.getValue();
            properties.numBoxes = tab.numBoxes.getValue();

            {
                setDebugObserverDatafield("numBoxes", properties.numBoxes);
                setDebugObserverDatafield("boxLayerName", properties.boxLayerName);
                sendDebugObserverBatch();
            }
        }
        if (tab.drawBoxes.wasClicked())
        {
            // Lock shared variables in methods running in separate threads
            // and pass them to functions. This way, the called functions do
            // not need to think about locking.
            std::scoped_lock lock(propertiesMutex);
            drawBoxes(properties, arviz);
        }
    }
    */


    /* (Requires the armarx::ArVizComponentPluginUser.)
    void
    Component::drawBoxes(const Component::Properties& p, viz::Client& arviz)
    {
        // Draw something in ArViz (requires the armarx::ArVizComponentPluginUser.
        // See the ArVizExample in RobotAPI for more examples.

        viz::Layer layer = arviz.layer(p.boxLayerName);
        for (int i = 0; i < p.numBoxes; ++i)
        {
            layer.add(viz::Box("box_" + std::to_string(i))
                      .position(Eigen::Vector3f(i * 100, 0, 0))
                      .size(20).color(simox::Color::blue()));
        }
        arviz.commit(layer);
    }
    */


    ARMARX_DECOUPLED_REGISTER_COMPONENT(Component);

}  /// namespace armarx::control::components::controller_creator
