/*
* This file is part of ArmarX.
*
* Copyright (C) 2011-2017, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
*
* ArmarX is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* ArmarX is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    ArmarX
* @author     Stefan Reither( stefan dot reither at kit dot edu)
* @date       2021
* @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
*             GNU General Public License
*/

#pragma once

#include <RobotAPI/components/units/RobotUnit/util/RtTiming.h>

#include "ErrorReporting.h"

namespace armarx::control::ethercat
{
    namespace timing
    {
        constexpr bool enableTimingInformation = false;

        namespace general
        {
            class Timer
            {
            public:
                void
                start()
                {
                    start_ = armarx::rtNow();
                }

                void
                end(const char* comment)
                {
                    GENERAL_INFO("%s - duration: %.3f ms",
                                 comment,
                                 (armarx::rtNow() - start_).toMilliSecondsDouble());
                }

                void
                end(const char* comment, double thresholdMS)
                {
                    const double diff = (armarx::rtNow() - start_).toMilliSecondsDouble();
                    if (diff >= thresholdMS)
                    {
                        GENERAL_INFO("%s - duration: %.3f ms", comment, diff);
                    }
                }

            private:
                IceUtil::Time start_;
            };
        } // namespace general

        namespace bus
        {
            class Timer
            {
            public:
                void
                start()
                {
                    start_ = armarx::rtNow();
                }

                void
                end(const std::uint64_t bin, const char* comment)
                {
                    BUS_INFO(bin,
                             "%s - duration: %.3f ms",
                             comment,
                             (armarx::rtNow() - start_).toMilliSecondsDouble());
                }

                void
                end(const std::uint64_t bin, const char* comment, double thresholdMS)
                {
                    const auto diff = (armarx::rtNow() - start_).toMilliSecondsDouble();
                    if (diff >= thresholdMS)
                    {
                        BUS_INFO(bin, "%s - duration: %.3f ms", comment, diff);
                    }
                }

            private:
                IceUtil::Time start_;
            };
        } // namespace bus

        namespace slave
        {
            class Timer
            {
            public:
                void
                start()
                {
                    start_ = armarx::rtNow();
                }

                void
                end(const SlaveIdentifier sid, const char* comment)
                {
                    SLAVE_INFO(sid,
                               "%s - duration: %.3f ms",
                               comment,
                               (armarx::rtNow() - start_).toMilliSecondsDouble());
                }

                void
                end(const SlaveIdentifier sid, const char* comment, double thresholdMS)
                {
                    const auto diff = (armarx::rtNow() - start_).toMilliSecondsDouble();
                    if (diff >= thresholdMS)
                    {
                        SLAVE_INFO(sid, "%s - duration: %.3f ms", comment, diff);
                    }
                }

            private:
                IceUtil::Time start_;
            };
        } // namespace slave
    } // namespace timing

#define GENERAL_TIMING_START(name)                                                                 \
    auto name = armarx::control::ethercat::timing::general::Timer();                               \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.start();                                                                              \
    }
#define GENERAL_TIMING_END_COMMENT(name, comment)                                                  \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.end(comment);                                                                         \
    }
#define GENERAL_TIMING_END(name) GENERAL_TIMING_END_COMMENT(name, #name)
#define GENERAL_TIMING_CEND_COMMENT(name, comment, thresholdMs)                                    \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.end(comment, thresholdMs);                                                            \
    }
#define GENERAL_TIMING_CEND(name, thresholdMs) GENERAL_TIMING_CEND_COMMENT(name, #name, thresholdMs)

    /////////////////////////////////////////////////////////////////////////////////////////////////

#define BUS_TIMING_START(name)                                                                     \
    auto name = armarx::control::ethercat::timing::bus::Timer();                                   \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.start();                                                                              \
    }

#define BUS_TIMING_END_COMMENT(name, bin, comment)                                                 \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.end(bin, comment);                                                                    \
    }
#define BUS_TIMING_END(name, bin) BUS_TIMING_END_COMMENT(name, bin, #name)
#define BUS_TIMING_CEND_COMMENT(name, bin, comment, thresholdMs)                                   \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.end(bin, comment, thresholdMs);                                                       \
    }
#define BUS_TIMING_CEND(name, bin, thresholdMs)                                                    \
    BUS_TIMING_CEND_COMMENT(name, bin, #name, thresholdMs)

    /////////////////////////////////////////////////////////////////////////////////////////////////

#define SLAVE_TIMING_START(name)                                                                   \
    auto name = armarx::control::ethercat::timing::slave::Timer();                                 \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.start();                                                                              \
    }
#define SLAVE_TIMING_END_COMMENT(sid, name, comment)                                               \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.end(sid, comment);                                                                    \
    }
#define SLAVE_TIMING_END(sid, name) SLAVE_TIMING_END_COMMENT(sid, name, #name)
#define SLAVE_TIMING_CEND_COMMENT(sid, name, comment, thresholdMs)                                 \
    if constexpr (armarx::control::ethercat::timing::enableTimingInformation)                      \
    {                                                                                              \
        name.end(sid, comment, thresholdMs);                                                       \
    }
#define SLAVE_TIMING_CEND(sid, name, thresholdMs)                                                  \
    SLAVE_TIMING_CEND_COMMENT(sid, name, #name, thresholdMs)

} // namespace armarx::control::ethercat
