/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ...
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <VirtualRobot/Robot.h>

#include <RobotAPI/components/units/RobotUnit/RobotUnit.h>
#include <RobotAPI/components/units/RobotUnit/NJointControllers/NJointController.h>
#include <RobotAPI/components/units/RobotUnit/ControlTargets/ControlTarget1DoFActuator.h>
#include <RobotAPI/components/units/RobotUnit/SensorValues/SensorValue1DoFActuator.h>

#include <armarx/control/common/FTSensor.h>
#include <armarx/control/common/control_law/KeypointsAdmittanceController.h>
#include <armarx/control/common/control_law/ControllerCommonInterface.h>
#include <armarx/control/njoint_controller/task_space/ControllerInterface.h>


namespace armarx::control::njoint_controller::task_space
{
    namespace common = armarx::control::common;
    namespace law = armarx::control::common::control_law;
    /**
    * @defgroup Library-NJointKeypointsAdmittanceController NJointKeypointsAdmittanceController
    * @ingroup Library-RobotUnit-NJointControllers
    * A description of the library NJointKeypointsAdmittanceController.
    *
    * @class NJointKeypointsAdmittanceController
    * @ingroup Library-NJointKeypointsAdmittanceController
    * @brief Brief description of class NJointKeypointsAdmittanceController.
    *
    * Detailed description of class NJointKeypointsAdmittanceController.
    */
    class NJointKeypointsAdmittanceController :
        virtual public NJointControllerWithTripleBuffer<law::KeypointsAdmittanceController::Config>,
        virtual public NJointKeypointsAdmittanceControllerInterface
    {
    public:
        using ConfigPtrT = NJointTaskspaceControllerConfigPtr;

        NJointKeypointsAdmittanceController(const RobotUnitPtr& robotUnit, const NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&);
        std::string getClassName(const Ice::Current&) const override;
        void rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) override;

        // NJointController interface
        std::string getKinematicChainName(const Ice::Current&) override;

        /// set control target
        void reconfigureController(const std::string &filename, const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void setControlParameters(const std::string&, const Eigen::VectorXf&, const Ice::Current&) override;
        void setKeypointsParameters(const std::string &name, const Eigen::VectorXf &value, const Ice::Current &) override;

        /// ft sensor
        void setForceTorqueBaseline(const Eigen::Vector3f&, const Eigen::Vector3f&, const Ice::Current&) override;
        void toggleGravityCompensation(const bool toggle, const Ice::Current &) override;
        void setTCPMass(Ice::Float, const Ice::Current&) override;
        void setTCPCoMInFTFrame(const Eigen::Vector3f&, const Ice::Current&) override;
        void calibrateFTSensor(const Ice::Current&) override;

    protected:
        virtual void additionalTask(){};
        void onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&) override;

    protected:
        /// devices
        std::vector<const SensorValue1DoFActuatorTorque*> torqueSensors;
        std::vector<const SensorValue1DoFActuatorVelocity*> velocitySensors;
        std::vector<const SensorValue1DoFActuatorPosition*> positionSensors;
        std::vector<ControlTarget1DoFActuatorTorque*> targets;


    protected:
        /// variables
        std::string configFileName;
        std::vector<std::string> jointNames;
        law::KeypointsAdmittanceController controller;
        common::FTSensor ftsensor;

        /// set buffers
        TripleBuffer<law::KeypointsAdmittanceController::Status> controlStatusBuffer;
        TripleBuffer<common::FTSensor::FTBufferData> ftSensorBuffer;

        std::atomic_bool rtFirstRun = true;
        std::string kinematicChainName;
        std::atomic_bool rtReady = false;

    protected:
        void onInitNJointController() override;
        void rtPreActivateController() override;
    };
} /// namespace armarx::control::njoint_controller::task_space
