/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ...
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <armarx/control/common/mp/KeypointsMP.h>
#include <armarx/control/common/mp/MPPool.h>
#include <armarx/control/njoint_controller/task_space/KeypointsImpedanceController.h>
#include <armarx/control/njoint_mp_controller/task_space/ControllerInterface.h>
#include <armarx/control/njoint_mp_controller/task_space/aron/NJointMPControllerConfig.aron.generated.h>



namespace armarx::control::njoint_mp_controller::task_space
{
    using namespace armarx::control::njoint_controller::task_space;
    using namespace armarx::control::common;


    struct Constraint
    {
        std::string type;
        Eigen::Vector3f pcaMean;
        Eigen::Matrix3f pcaComponents;
        Eigen::Matrix3f localFrameRotMat;
        Eigen::Vector3f localFrameOrigin;
        Eigen::VectorXf vmpStart;

        int priority = 0;
        Eigen::MatrixXf density_mu;
        Eigen::VectorXf density_weights;
        float density_std_dev = 0.0f;
        float spatial_scale = 0.0;

        float pm_scale = 1.0;
        Eigen::Vector3f pm_init_proj_point;
        float pm_init_proj_value = 0.0;
        Eigen::Vector3f pm_init_tangent_vec;

    };
    using ConstraintPtr = std::shared_ptr<Constraint>;


    struct KVILImpedanceMPConfig:
            virtual public control_law::KeypointsImpedanceController::Config
    {
        mp::MP::MPListConfig mpConfig;
        std::vector<Constraint> constraintList;
    };


    /**
    * @defgroup Library-NJointKVILImpedanceMPController NJointKVILImpedanceMPController
    * @ingroup Library-RobotUnit-NJointControllers
    * A description of the library NJointKVILImpedanceMPController.
    *
    * @class NJointKVILImpedanceMPController
    * @ingroup Library-NJointKVILImpedanceMPController
    * @brief Brief description of class NJointKVILImpedanceMPController.
    *
    * Detailed description of class NJointKVILImpedanceMPController.
    */
    class NJointKVILImpedanceMPController :
            virtual public NJointKeypointsImpedanceController
            , virtual public mp::MPPool
            , virtual public NJointKVILImpedanceMPControllerInterface
    {
    public:
        using AronDTO = arondto::KVILImpedanceMPConfig;
        using BO = KVILImpedanceMPConfig;

        NJointKVILImpedanceMPController(const RobotUnitPtr& robotUnit, const NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&);

        std::string getClassName(const Ice::Current&) const override;
        void updateConfig(const ::armarx::aron::data::dto::DictPtr& dto,
                          const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void rtPreActivateController() override;
        void additionalTask() override;

//        void configureConstraints(const std::vector<Constraint>&c);

        void setCurveProjection(const Eigen::Vector3f& curve_proj_point_, float proj_value_, const Eigen::Vector3f& curve_vec_, const Ice::Current &) override;
        Eigen::Vector3f getCurveKeypointPosition(const Ice::Current &) override;

    private:
        std::map<std::string, ConstraintPtr> constraints;
        int n_constraints = 0;

        struct NonlinearManifoldInfoStruct
        {
            Eigen::Vector3f pm_proj_point;
            float pm_proj_point_to_mean_distance = 100.0f;
            Eigen::Vector3f pm_tangent_vec;
        };

        struct p2CStruct
        {
            Eigen::Vector3f curve_kpt_position;
        };

        TripleBuffer<NonlinearManifoldInfoStruct> nl_mani_buffer;
        TripleBuffer<p2CStruct> kpt_status_buffer;

    protected:
        TripleBuffer<BO> bufferUserToAdditionalTask;
        TripleBuffer<BO> bufferUserToRt;
    };
}
